package jp.gr.java_conf.abagames.pavg;

import java.io.*;
import java.util.*;

/**
 * PAVG Converter boot class
 */
public class PavgConverter {
  String srcFile, dstFile;

  public PavgConverter(String srcFile, String dstFile) {
    this.srcFile = srcFile; this.dstFile = dstFile;
  }

  private void deleteDstFile() {
    new File(dstFile).delete();
  }

  private final String PALM_ENCODING = "SJIS";

  /**
   * write string to database
   */
  private void writeString(OutputStream output, String str, int maxLength)
   throws UnsupportedEncodingException, IOException {
    byte[] strBytes;
    strBytes = str.getBytes(PALM_ENCODING);
    int lgt = strBytes.length;
    if ( lgt >= maxLength ) lgt = maxLength-1;
    for ( int i=0 ; i<lgt ; i++ ) {
      output.write(strBytes[i]);
    }
    for ( int i=lgt ; i<maxLength ; i++ ) {
      output.write((byte)'\0');
    }
  }

  /**
   * write 4 bytes to database
   */
  private void writeLong(OutputStream output, long data) throws IOException {
    output.write((int)((data&0xff000000)>>24));
    output.write((int)((data&0x00ff0000)>>16));
    output.write((int)((data&0x0000ff00)>>8 ));
    output.write((int)((data&0x000000ff)    ));
  }

  private void writeLong(OutputStream output, String data) throws IOException {
    byte[] strBytes = new byte[4];
    strBytes = data.getBytes();
    for ( int i=0 ; i<4 ; i++ ) {
      output.write(strBytes[i]);
    }
  }

  /**
   * write 2 bytes to database
   */
  private void writeShort(OutputStream output, int data) throws IOException {
    output.write((int)((data&0x0000ff00)>>8 ));
    output.write((int)((data&0x000000ff)    ));
  }

  private final int VERSION_NUM = 100;
  private final String PAVG_CID = "PAVG";

  private final int TITLE_LENGTH = 50;
  private final int COPYRIGHT_LENGTH = 34;
  private final int MESSAGE_LENGTH = 194;
  private final int CHOICES_LENGTH = 34;

  public void convert() throws IOException {
    BufferedReader reader
      = new BufferedReader(new FileReader(srcFile));
    //BufferedReader reader = new BufferedReader(
    //  new InputStreamReader(new FileInputStream(srcFile), "JISAutoDetect"));
    BufferedOutputStream database
      = new BufferedOutputStream(new FileOutputStream(dstFile));
    CSVStreamTokenizer enum = new CSVStreamTokenizer(reader);

    String title = enum.nextToken();
    if ( title == null ) {
      database.close();
      reader.close();
      deleteDstFile();
      throw new IOException("Scenario file format error: no title found.");
    }
    System.out.println("Convert: " + title + ".");

    // write database header
    String dbName = new String(title);
    if ( dbName.length() > 32-6 ) {
      dbName = dbName.substring(0, 32-6);
    }
    dbName += "-" + PAVG_CID;
    writeString(database, dbName , 32);
    writeShort(database, 0);
    writeShort(database, VERSION_NUM);
    // date field
    Date nowDate = new Date();
    Calendar baseCal = Calendar.getInstance();
    baseCal.set(1904, 0, 1);
    Date baseDate = baseCal.getTime();
    long crDate = (nowDate.getTime() - baseDate.getTime())/1000;
    writeLong(database, crDate);
    writeLong(database, crDate);
    writeLong(database, 0);

    writeLong(database, 0);
    writeLong(database, 0);
    writeLong(database, 0);
    writeLong(database, "DATA");
    writeLong(database, PAVG_CID);
    writeLong(database, 0);
    writeLong(database, 0);
    writeShort(database, 1);

    // write record list
    writeLong(database, 86);
    writeShort(database, 0);
    writeShort(database, 0);

    // write record
    writeString(database, title, TITLE_LENGTH);
    String copyRight = enum.nextToken();
    if ( copyRight == null ) {
      database.close();
      reader.close();
      deleteDstFile();
      throw new IOException("Scenario file format error: no copyright found.");
    }
    writeString(database, copyRight, COPYRIGHT_LENGTH);
    while ( !enum.isSeparatedByLineSeparator() ) enum.nextToken();
    for ( ; ; ) {
      String idxStr = enum.nextToken();
      if ( idxStr == null ) break;
      int idx;
      try {
        writeShort(database, idx = new Integer(idxStr.trim()).intValue());
      } catch ( NumberFormatException e ) {
        database.close();
        reader.close();
        deleteDstFile();
        throw new IOException("Scenario file format error: invalid index number.");
      }
      System.out.print("[" + idx + "] ");
      String msg = enum.nextToken();
      if ( msg == null ) {
        database.close();
        reader.close();
        deleteDstFile();
        throw new IOException("Scenario file format error: no message field.");
      }
      writeString(database, msg, MESSAGE_LENGTH);
      boolean eof = false;
      boolean sprtLn = enum.isSeparatedByLineSeparator();
      for ( int i=0 ; i<4 ; i++ ) {
        String chMsg;
        if ( !sprtLn ) {
          chMsg = enum.nextToken();
          if ( chMsg == null ) {
            eof = true;
            chMsg = new String();
          } else {
            sprtLn = enum.isSeparatedByLineSeparator();
          }
        } else {
          chMsg = new String();
        }
        writeString(database, chMsg, CHOICES_LENGTH);
        String chIdx;
        if ( !sprtLn ) {
          chIdx = enum.nextToken();
          if ( chIdx == null ) {
            eof = true;
            chIdx = new String();
          } else {
            sprtLn = enum.isSeparatedByLineSeparator();
          }
        } else {
          chIdx = new String();
        }
        int chIdxNum;
        try {
          chIdxNum = new Integer(chIdx.trim()).intValue();
        } catch ( NumberFormatException e ) {
          chIdxNum = -1;
        }
        writeShort(database, chIdxNum);
      }
      if ( eof ) break;
    }
    database.close();
    reader.close();
    System.out.println();
    System.out.println("Finished.");
  }

  public static void main(String[] args) {
    PavgConverter converter = new PavgConverter(args[0], args[1]);
    try {
      converter.convert();
    } catch ( Exception e ) {
      System.out.println();
      System.out.println("Convert failed.");
      e.printStackTrace();
    }
  }
}
