/**
 * Palm Adventure
 * (c) Kenta Cho (cs8k-cyu@asahi-net.or.jp)
 * http://www.asahi-net.or.jp/~cs8k-cyu/
 *
 * pavg.c(main routine)
 */

#include <PalmOS.h>
#include "pavg.h"


/********************************
  scenario database management 
 ********************************/
static DmOpenRef gDB;

static Scenario* scenario;
static MemHandle scenarioRecord = NULL;

static int* itemTable;
static int itemNum;
static MemHandle itemTableHandle;
static int maxIndex;

static MemHandle getScenarioData(ScenarioDatabase scDb) {
  int i;
  LocalID lid = DmFindDatabase(scDb.cardNo, scDb.name);
  maxIndex = -1;
  if ( lid == 0 ) {
    ErrDisplay("no scenario data.");
    return NULL;
  }
  gDB = DmOpenDatabase(scDb.cardNo, lid, dmModeReadOnly);
  if ( gDB == 0 ) {
    ErrDisplay("open scenario data failed.");
    return NULL;
  }
  scenarioRecord = DmQueryRecord(gDB, 0);
  if ( scenarioRecord == 0 ) {
    ErrDisplay("open scenario data failed.");
    return NULL;
  }
  scenario = MemHandleLock(scenarioRecord);
  itemNum = 
    (MemHandleSize(scenarioRecord)-TITLE_LENGTH-COPYRIGHT_LENGTH)/sizeof(Item);
  for ( i=0 ; i<itemNum ; i++ ) {
    if ( scenario->item[i].index > maxIndex ) {
      maxIndex = scenario->item[i].index;
    }
  }
  if ( maxIndex < 0 ) {
    ErrDisplay("no contents.");
    return NULL;
  }
  itemTableHandle = MemHandleNew((maxIndex+1)*sizeof(int));
  if ( itemTableHandle == NULL ) {
    ErrDisplay("no memory.");
    return NULL;
  }
  itemTable = MemHandleLock(itemTableHandle);
  for ( i=0 ; i<maxIndex ; i++ ) {
    itemTable[i] = -1;
  }
  for ( i=0 ; i<itemNum ; i++ ) {
    itemTable[scenario->item[i].index] = i;
  }
  MemHandleUnlock(itemTableHandle);
  MemHandleUnlock(scenarioRecord);
  return scenarioRecord;
}

static void closeScenarioData() {
  if ( itemTableHandle != NULL ) {
    MemHandleFree(itemTableHandle);
    itemTableHandle = NULL;
  }
  if ( scenarioRecord != NULL ) {
    DmCloseDatabase(gDB);
    scenarioRecord = NULL;
  }
}

static void deleteScenarioData(ScenarioDatabase scDb) {
  LocalID lid = DmFindDatabase(scDb.cardNo, scDb.name);
  if ( lid == 0 ) {
    ErrDisplay("no scenario data.");
    return;
  }
  DmDeleteDatabase(scDb.cardNo, lid);
}


/********************************
  select scenario
 ********************************/
static GameStatus gameStatus;

MemHandle scDbHandle;
int scenarioNum;

#define DATABASE_TYPE 'DATA'

static void getScenarioList() {
  int i;
  ScenarioDatabase* scDb;
  DmSearchStateType dss;
  Boolean fs = TRUE;
  UInt16 cardNo;
  LocalID lid;
  scenarioNum = 0;
  for ( ; ; ) {
    if ( 0 != DmGetNextDatabaseByTypeCreator
	(fs, &dss, DATABASE_TYPE, PAVG_CID, FALSE, &cardNo, &lid) ) break;
    scenarioNum++;
    fs = FALSE;
  }
  if ( scenarioNum == 0 ) return;
  scDbHandle = MemHandleNew(scenarioNum*sizeof(ScenarioDatabase));
  scDb = MemHandleLock(scDbHandle);
  fs = TRUE;
  for ( i=0 ; i<scenarioNum ; i++ ) {
    char name[35];
    DmGetNextDatabaseByTypeCreator
      (fs, &dss, DATABASE_TYPE, PAVG_CID, FALSE, &cardNo, &lid);
    DmDatabaseInfo(cardNo, lid, name, NULL, NULL, NULL, NULL, NULL, NULL, 
		   NULL, NULL, NULL, NULL);
    StrCopy(scDb[i].name, name);
    scDb[i].cardNo = cardNo;
    fs = FALSE;
  }
  MemHandleUnlock(scDbHandle);
}

/**
 * draw scenario names to list
 */
static void scenarioListDrawFunc(UInt16 itemNum, RectangleType *bounds, 
				 Char **data) {
  ScenarioDatabase* scDb = MemHandleLock(scDbHandle);  
  char* s = scDb[itemNum].name;
  int tx = bounds->topLeft.x, ty = bounds->topLeft.y;
  int ex = bounds->extent.x, ey = bounds->extent.y;
  WinDrawChars(s, StrLen(s)-5, tx+6, ty);
  WinDrawLine(tx, ty+ey-1, tx+ex, ty+ey-1);
  WinDrawLine(tx, ty+ey-1, tx, ty+ey-5);
  WinDrawLine(tx, ty+ey-5, tx+4, ty+ey-1);
  MemHandleUnlock(scDbHandle);
}
  
/**
 * init scenario names list 
 */
static void initScenarioList() {
  FormType *frm = FrmGetActiveForm();
  ListType *list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, SCENARIO_LIST));
  LstSetListChoices(list, NULL, scenarioNum);
  LstSetDrawFunction(list, (ListDrawDataFuncPtr)scenarioListDrawFunc);
  LstSetSelection(list, -1);
}

static void gotoScenarioForm() {
  gameStatus.itemIndex = -1;
  FrmGotoForm(SCENARIO_FORM);
}


/********************************
  game management
 ********************************/
static void resetGame() {
  gameStatus.itemIndex = 0;
}

static void initGame() {
  gameStatus.bookmark = -1;
  resetGame();
}

/**
 * show about box
 */
static void showAboutBox() {
  MemHandle verH;
  Char* verStr;
  verH = DmGetResource(verRsc, VERSION);
  verStr = MemHandleLock(verH);
  FrmCustomAlert(ABOUT_DIALOG, verStr, scenario->copyright, NULL);
  MemHandleUnlock(verH);
  DmReleaseResource(verH);
}

/**
 * set field text
 */
static FieldType* setFieldTextFromHandle(UInt16 fieldID, MemHandle txtH) {
  MemHandle oldTxtH;
  FormType* frm = FrmGetActiveForm();
  FieldType* fldP;
  fldP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  ErrNonFatalDisplayIf(!fldP, "missing field.");
  oldTxtH = (MemHandle)FldGetTextHandle(fldP);
  FldSetTextHandle(fldP, txtH);
  FldDrawField(fldP);
  if ( oldTxtH ) MemHandleFree(oldTxtH);
  return fldP;
}

static FieldType* setFieldText(UInt16 fieldID, char* str) {
  MemHandle txtH = MemHandleNew(StrLen(str)+1);
  if ( !txtH ) return NULL;
  StrCopy(MemHandleLock(txtH), str);
  MemHandleUnlock(txtH);
  return setFieldTextFromHandle(fieldID, (MemHandle)txtH);
}

/**
 * draw bitmap
 */
static void drawBitmap(int id, int x, int y) {
  MemHandle objectBitmapHandles;
  BitmapType *objectBitmapPtr;
  objectBitmapHandles = DmGetResource(bitmapRsc, id);
  objectBitmapPtr = MemHandleLock(objectBitmapHandles);
  WinDrawBitmap(objectBitmapPtr, x, y);
  MemPtrUnlock(objectBitmapPtr);
  DmReleaseResource(objectBitmapHandles);
}

/**
 * draw index number
 */
#define NUMBER_Y 20

static void clearNumber() {
  RectangleType rct;
  rct.topLeft.x = 0;
  rct.topLeft.y = NUMBER_Y;
  rct.extent.x = 30;
  rct.extent.y = 16;
  WinEraseRectangle(&rct, 0);
}

static void drawNumber(int num) {
  clearNumber();
  if ( num < 0 ) return;
  if ( num >= 100 ) {
    drawBitmap(ZERO_BITMAP+(num%10), 20, NUMBER_Y);
    num /= 10;
    drawBitmap(ZERO_BITMAP+(num%10), 10, NUMBER_Y);
    num /= 10;
    drawBitmap(ZERO_BITMAP+(num%10), 0, NUMBER_Y);
  } else if ( num >= 10 ) {
    drawBitmap(ZERO_BITMAP+(num%10), 15, NUMBER_Y);
    num /= 10;
    drawBitmap(ZERO_BITMAP+(num%10), 5, NUMBER_Y);
  } else {
    drawBitmap(ZERO_BITMAP+num, 10, NUMBER_Y);
  }
}

/**
 * draw current item(index number, message and choices)
 * (when this function is called, scenario pointer should be valid)
 */
static char nextIndexStr[CHOICES_NUM][5];

static int drawItem() {
  int idx = gameStatus.itemIndex;
  int i;
  FormType *frm = FrmGetActiveForm();
  MenuEraseStatus(NULL);
  drawNumber(scenario->item[idx].index);
  setFieldText(MESSAGE_FIELD, scenario->item[idx].msg);
  for ( i=0 ; i<CHOICES_NUM ; i++ ) {
    Choices *ch = &(scenario->item[idx].choices[i]);
    UInt16 fieldIdx = FrmGetObjectIndex(frm, CHOICES1_FIELD+i);
    UInt16 buttonIdx = FrmGetObjectIndex(frm, CHOICES1_BUTTON+i);
    FrmHideObject(frm, fieldIdx);
    FrmHideObject(frm, buttonIdx);
    if ( ch->nextIndex > 0 ) {
      ControlType *button;
      setFieldText(CHOICES1_FIELD+i, ch->msg);
      StrIToA(nextIndexStr[i], (ch->nextIndex)%1000);
      button = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, CHOICES1_BUTTON+i));
      CtlSetLabel(button, nextIndexStr[i]);
      FrmShowObject(frm, fieldIdx);
      FrmShowObject(frm, buttonIdx);
    }
  }
}

/**
 * start game(set scenario title)
 */
static void startGame() {
  FormType *frm = FrmGetActiveForm();
  scenario = MemHandleLock(scenarioRecord);
  FrmSetTitle(frm, scenario->title);
  drawItem();
  MemHandleUnlock(scenarioRecord);
}

/**
 * choices button is pushed and go to another item
 */
static void choicesButtonPushed(int num) {
  int idx = gameStatus.itemIndex;
  int nextIndex, nextItemIndex;
  scenario = MemHandleLock(scenarioRecord);
  itemTable = MemHandleLock(itemTableHandle);
  nextIndex = scenario->item[idx].choices[num].nextIndex;
  if ( nextIndex > maxIndex ) {
    FrmAlert(NO_INDEX_DIALOG);
    MemHandleUnlock(itemTableHandle);
    MemHandleUnlock(scenarioRecord);
    return;
  }
  nextItemIndex = itemTable[nextIndex];
  if ( nextItemIndex < 0 ) {
    FrmAlert(NO_INDEX_DIALOG);
    MemHandleUnlock(itemTableHandle);
    MemHandleUnlock(scenarioRecord);
    return;
  }
  gameStatus.itemIndex = nextItemIndex;
  MemHandleUnlock(itemTableHandle);
  drawItem();
  MemHandleUnlock(scenarioRecord);
}

/**
 * set bookmark to current item index
 */
static void setBookmark() {
  gameStatus.bookmark = gameStatus.itemIndex;
}

/**
 * go to bookmarked item index
 */
static void gotoBookmark() {
  if ( gameStatus.bookmark < 0 ) {
    FrmAlert(NO_BOOKMARK_DIALOG);
    return;
  }
  gameStatus.itemIndex = gameStatus.bookmark;
  MemHandleLock(scenarioRecord);
  drawItem();
  MemHandleUnlock(scenarioRecord);
}

static void gotoGameForm() {
  FrmGotoForm(GAME_FORM);
}


/********************************
  application main routine
 ********************************/
/**
 * event hanlder for scenario selection form
 */
static Boolean ScenarioFormHandleEvent(EventPtr event) {
  int idx;
  ScenarioDatabase* scDb;
  ListType *list;
  Boolean handled = FALSE;
  FormType *frm = FrmGetActiveForm();
  switch ( event->eType ) {
  case ctlSelectEvent:
    handled = TRUE;
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, SCENARIO_LIST));
    idx = LstGetSelection(list);
    if ( idx == -1 ) break;
    if ( event->data.ctlSelect.controlID == START_BUTTON ) {
      /* start button pushed */
      scDb = MemHandleLock(scDbHandle);
      if ( NULL == getScenarioData(scDb[idx]) ) {
	MemHandleUnlock(scDbHandle);
	break;
      }
      gameStatus.scenarioDatabase = scDb[idx];
      MemHandleUnlock(scDbHandle);
      initGame();
      gotoGameForm();
      break;
    } else if ( event->data.ctlSelect.controlID == DELETE_BUTTON ) {
      /* delete button pushed */
      scDb = MemHandleLock(scDbHandle);
      deleteScenarioData(scDb[idx]);
      MemHandleUnlock(scDbHandle);
      getScenarioList();
      gotoScenarioForm();
    }
    break;
  case frmOpenEvent:
    initScenarioList();
    FrmDrawForm(frm);
    handled = TRUE;
    break;
  }
  return handled;
}

/**
 * event hanlder for game form
 */
static Boolean GameFormHandleEvent(EventPtr event) {
  Int16 px, py;
  Boolean pd;
  Boolean handled = FALSE;
  switch ( event->eType ) {
  case ctlSelectEvent:
    /* choices button pushed */
    switch ( event->data.ctlSelect.controlID ) {
    case CHOICES1_BUTTON:
      choicesButtonPushed(0);
      break;
    case CHOICES2_BUTTON:
      choicesButtonPushed(1);
      break;
    case CHOICES3_BUTTON:
      choicesButtonPushed(2);
      break;
    case CHOICES4_BUTTON:
      choicesButtonPushed(3);
      break;
    }
    handled = TRUE;
    break;
  case menuEvent:
    /* menu selected */
    switch ( event->data.menu.itemID ) {
    case SCENARIO_MENU:
      closeScenarioData();
      gotoScenarioForm();
      break;
    case RESTART_MENU:
      resetGame();
      startGame();
      break;
    case SET_MENU:
      setBookmark();
      break;
    case GOTO_MENU:
      gotoBookmark();
      break;
    case ABOUT_MENU:
      showAboutBox();
      break;
    }
    handled = TRUE;
    break;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    startGame();
    handled = TRUE;
    break;
  }
  return handled;
}

/**
 * activate forms
 */
static Boolean ApplicationHandleEvent(EventPtr event) {
  FormType *frm;
  Int16 formID;
  Boolean handled = FALSE;

  if ( event->eType == frmLoadEvent ) {
    formID = event->data.frmLoad.formID;
    frm = FrmInitForm(formID);
    FrmSetActiveForm(frm);
    switch ( formID ) {
    case GAME_FORM:
      FrmSetEventHandler(frm, GameFormHandleEvent);
      break;
    case SCENARIO_FORM:
      FrmSetEventHandler(frm, ScenarioFormHandleEvent);
      break;
    }
    handled = TRUE;
  }
  return handled;
}

static void EventLoop(void) {
  EventType event;
  UInt16 error;
  do {
    EvtGetEvent(&event, evtWaitForever);
    if ( !SysHandleEvent(&event) ) {
      if ( !MenuHandleEvent(0, &event, &error) ) {
	if ( !ApplicationHandleEvent(&event) ) {
	  FrmDispatchEvent(&event);
	}
      }
    }
  } while ( event.eType != appStopEvent );
}

static void StartApplication(void) {
  Boolean exist;
  getScenarioList();
  exist = PrefGetAppPreferencesV10
    (PAVG_CID, PAVG_VER, &gameStatus, sizeof(gameStatus));
  if ( exist ) {
    if ( gameStatus.itemIndex >= 0 ) {
      if ( NULL != getScenarioData(gameStatus.scenarioDatabase) ) {
	if ( gameStatus.itemIndex < itemNum ) {
	  gotoGameForm();
	  return;
	}
      }
    } 
  }
  initGame();
  gotoScenarioForm();
}

static void StopApplication(void) {
  PrefSetAppPreferencesV10
    (PAVG_CID, PAVG_VER, &gameStatus, sizeof(gameStatus));
  closeScenarioData();
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags) {
  if ( cmd != sysAppLaunchCmdNormalLaunch ) return 0;
  StartApplication();
  EventLoop();
  StopApplication();
  return 0;
}
